@@ -5,8 +5,8 @@
 ;; Author: Sergey Kostyaev <sskostyaev@gmail.com>
 ;; URL: http://github.com/s-kostyaev/ellama
 ;; Keywords: help local tools
-;; Package-Requires: ((emacs "28.1") (llm "0.6.0") (spinner "1.7.4"))
-;; Version: 0.6.0
+;; Package-Requires: ((emacs "28.1") (llm "0.6.0") (spinner "1.7.4") (dash "2.19.1"))
+;; Version: 0.7.7
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 ;; Created: 8th Oct 2023
 
@@ -38,6 +38,7 @@
 (require 'json)
 (require 'llm)
 (require 'spinner)
+(require 'dash)
 (eval-when-compile (require 'rx))
 
 (defgroup ellama nil
@@ -54,7 +55,7 @@
   :group 'ellama
   :type 'string)
 
-(defcustom ellama-nick-prefix "##"
+(defcustom ellama-nick-prefix "**"
   "User and assistant nick prefix in logs."
   :group 'ellama
   :type 'string)
@@ -253,6 +254,42 @@ It should be a function with single argument generated text string."
   :group 'ellama
   :type 'function)
 
+(defcustom ellama-major-mode 'org-mode
+  "Major mode for ellama commands."
+  :group 'ellama
+  :type 'symbol)
+
+(defcustom ellama-long-lines-length 100
+  "Long lines length for fill paragraph call.
+Too low value can break generated code by splitting long comment lines."
+  :group 'ellama
+  :type 'integer)
+
+(defcustom ellama-session-auto-save t
+  "Automatically save ellama sessions if set."
+  :group 'ellama
+  :type 'boolean)
+
+(define-minor-mode ellama-session-mode
+  "Minor mode for ellama session buffers."
+  :interactive nil
+  (if ellama-session-mode
+      (progn
+        (add-hook 'after-save-hook 'ellama--save-session nil t)
+        (add-hook 'kill-buffer-hook 'ellama--session-deactivate nil t))
+    (remove-hook 'kill-buffer-hook 'ellama--session-deactivate)
+    (remove-hook 'after-save-hook 'ellama--save-session)
+    (ellama--session-deactivate)))
+
+(define-minor-mode ellama-request-mode
+  "Minor mode for ellama buffers with active request to llm."
+  :interactive nil
+  :keymap '(([remap keyboard-quit] . ellama--cancel-current-request-and-quit))
+  (if ellama-request-mode
+      (add-hook 'kill-buffer-hook 'ellama--cancel-current-request nil t)
+    (remove-hook 'kill-buffer-hook 'ellama--cancel-current-request)
+    (ellama--cancel-current-request)))
+
 (defvar-local ellama--change-group nil)
 
 (defvar-local ellama--current-request nil)
@@ -270,6 +307,63 @@ It should be a function with single argument generated text string."
   ;; Trim left first as `string-trim' trims from the right and ends up deleting all the code.
   (string-trim-right (string-trim-left text ellama--code-prefix) ellama--code-suffix))
 
+(defun ellama--fill-string (s)
+  "Fill string S."
+  (with-temp-buffer
+    (emacs-lisp-mode)
+    (insert s)
+    (fill-region (point-min) (point-max))
+    (buffer-substring (point-min) (point-max))))
+
+(defun ellama--fill-long-lines (text)
+  "Fill long lines only in TEXT."
+  (--> text
+       (split-string it "\n")
+       (-map (lambda (el)
+	       (if (> (length el)
+		      ellama-long-lines-length)
+		   (ellama--fill-string el)
+		 el)) it)
+       (string-join it "\n")))
+
+(defun ellama--translate-markdown-to-org-filter (text)
+  "Filter to translate code blocks from markdown syntax to org syntax in TEXT.
+This filter contains only subset of markdown syntax to be good enough."
+  (->> text
+       ;; code blocks
+       (replace-regexp-in-string "^```\\(.+\\)$" "#+BEGIN_SRC \\1")
+       (replace-regexp-in-string "^<!-- language: \\(.+\\) -->\n```" "#+BEGIN_SRC \\1")
+       (replace-regexp-in-string "^```$" "#+END_SRC")
+       ;; lists
+       (replace-regexp-in-string "^\\* " "+ ")
+       ;; bold
+       (replace-regexp-in-string "\\*\\*\\(.+?\\)\\*\\*" "*\\1*")
+       (replace-regexp-in-string "__\\(.+?\\)__" "*\\1*")
+       (replace-regexp-in-string "<b>\\(.+?\\)</b>" "*\\1*")
+       ;; italic
+       (replace-regexp-in-string "_\\(.+?\\)_" "/\\1/")
+       (replace-regexp-in-string "<i>\\(.+?\\)</i>" "/\\1/")
+       ;; inline code
+       (replace-regexp-in-string "`\\(.+?\\)`" "~\\1~")
+       ;; underlined
+       (replace-regexp-in-string "<u>\\(.+?\\)</u>" "_\\1_")
+       ;; strikethrough
+       (replace-regexp-in-string "~~\\(.+?\\)~~" "+\\1+")
+       (replace-regexp-in-string "<s>\\(.+?\\)</s>" "+\\1+")
+       ;; headings
+       (replace-regexp-in-string "^# " "* ")
+       (replace-regexp-in-string "^## " "** ")
+       (replace-regexp-in-string "^### " "*** ")
+       (replace-regexp-in-string "^#### " "**** ")
+       (replace-regexp-in-string "^##### " "***** ")
+       (replace-regexp-in-string "^###### " "***** ")
+       ;; badges
+       (replace-regexp-in-string "\\[\\!\\[.*?\\](\\(.*?\\))\\](\\(.*?\\))" "[[\\2][file:\\1]]")
+       ;;links
+       (replace-regexp-in-string "\\[\\(.*?\\)\\](\\(.*?\\))" "[[\\2][\\1]]")
+       ;; filling long lines
+       (ellama--fill-long-lines)))
+
 (defcustom ellama-enable-keymap t
   "Enable or disable Ellama keymap."
   :type 'boolean
@@ -281,7 +375,7 @@ It should be a function with single argument generated text string."
 	   ;; If ellama-enable-keymap is nil, remove the key bindings
 	   (define-key global-map (kbd ellama-keymap-prefix) nil))))
 
-(defcustom ellama-session-file-extension "md"
+(defcustom ellama-session-file-extension "org"
   "File extension for saving ellama session."
   :type 'string
   :group 'ellama)
@@ -318,7 +412,8 @@ PROMPT is a variable contains last prompt in this session."
 
 (defun ellama-generate-name (provider action prompt)
   "Generate name for ellama ACTION by PROVIDER according to PROMPT."
-  (let ((prompt-words (split-string prompt)))
+  (let* ((cleaned-prompt (replace-regexp-in-string "/" "_" prompt))
+         (prompt-words (split-string cleaned-prompt)))
     (string-join
      (flatten-tree
       (list (split-string (format "%s" action) "-")
@@ -329,8 +424,10 @@ PROMPT is a variable contains last prompt in this session."
 	    (format "(%s)" (llm-name provider))))
      " ")))
 
-(defun ellama-new-session (provider prompt)
-  "Create new ellama session with provided PROVIDER and PROMPT and unique id."
+(defun ellama-new-session (provider prompt &optional ephemeral)
+  "Create new ellama session with unique id.
+Provided PROVIDER and PROMPT will be used in new session.
+If EPHEMERAL non nil new session will not be associated with any file."
   (let* ((name (ellama-generate-name provider 'ellama prompt))
 	 (count 1)
 	 (name-with-suffix (format "%s %d" name count))
@@ -340,35 +437,48 @@ PROMPT is a variable contains last prompt in this session."
 		 (setq count (+ count 1))
 		 (setq name-with-suffix (format "%s %d" name count)))
 	       name-with-suffix))
-	 (file-name (file-name-concat
-		     ellama-sessions-directory
-		     (concat id "." ellama-session-file-extension)))
-	 (session (make-ellama-session :id id :provider provider :file file-name))
-	 (buffer (progn
-		   (make-directory ellama-sessions-directory t)
-		   (find-file-noselect file-name))))
+	 (file-name (when (and (not ephemeral)
+			       ellama-session-auto-save)
+		      (file-name-concat
+		       ellama-sessions-directory
+		       (concat id "." ellama-session-file-extension))))
+	 (session (make-ellama-session
+		   :id id :provider provider :file file-name))
+	 (buffer (if file-name
+		     (progn
+		       (make-directory ellama-sessions-directory t)
+		       (find-file-noselect file-name))
+		   (get-buffer-create id))))
     (setq ellama--current-session-id id)
     (puthash id buffer ellama--active-sessions)
     (with-current-buffer buffer
-      (setq ellama--current-session session))
+      (funcall ellama-major-mode)
+      (setq ellama--current-session session)
+      (ellama-session-mode +1))
     session))
 
-(defun ellama--cancel-current-request (&rest _)
+(defun ellama--cancel-current-request ()
   "Cancel current running request."
   (when ellama--current-request
-    (llm-cancel-request ellama--current-request)))
+    (llm-cancel-request ellama--current-request)
+    (setq ellama--current-request nil)))
 
-(advice-add #'keyboard-quit :before #'ellama--cancel-current-request)
+(defun ellama--cancel-current-request-and-quit ()
+  "Cancel the current request and quit."
+  (interactive)
+  (ellama--cancel-current-request)
+  (keyboard-quit))
 
-(defun ellama--session-deactivate (&rest _)
+(defun ellama--session-deactivate ()
   "Deactivate current session."
-  (when ellama--current-session
-    (let ((id (ellama-session-id ellama--current-session)))
+  (ellama--cancel-current-request)
+  (when-let* ((session ellama--current-session)
+              (id (ellama-session-id session)))
+    (when (string= (buffer-name)
+                   (buffer-name (ellama-get-session-buffer id)))
+      (remhash id ellama--active-sessions)
       (when (equal ellama--current-session-id id)
-	(setq ellama--current-session-id nil))
-      (remhash id ellama--active-sessions))))
-
-(advice-add #'kill-buffer :before #'ellama--session-deactivate)
+	(setq ellama--current-session-id nil)))))
 
 (defun ellama--get-session-file-name (file-name)
   "Get ellama session file name for FILE-NAME."
@@ -380,19 +490,14 @@ PROMPT is a variable contains last prompt in this session."
 	   (concat "." base-name ".session.el"))))
     session-file-name))
 
-(defun ellama--save-session (&rest _)
+(defun ellama--save-session ()
   "Save current ellama session."
   (when ellama--current-session
     (let* ((session ellama--current-session)
 	   (file-name (ellama-session-file session))
 	   (session-file-name (ellama--get-session-file-name file-name)))
-      (with-current-buffer (find-file-noselect session-file-name)
-	(delete-region (point-min) (point-max))
-	(insert (concat "(setq ellama--current-session " (prin1-to-string session)")"))
-	(save-buffer)
-	(kill-buffer)))))
-
-(advice-add #'save-buffer :before #'ellama--save-session)
+      (with-temp-file session-file-name
+	(insert (prin1-to-string session))))))
 
 ;;;###autoload
 (defun ellama-load-session ()
@@ -414,12 +519,33 @@ PROMPT is a variable contains last prompt in this session."
 	      (buffer (find-file-noselect file-name))
 	      (session-buffer (find-file-noselect session-file-name)))
     (with-current-buffer session-buffer
-      (goto-char (point-min)))
+      (goto-char (point-min))
+      ;; old sessions support
+      (when (string= "(setq "
+		     (buffer-substring-no-properties 1 7))
+	(goto-char (point-min))
+	;; skip "("
+	(forward-char)
+	;; skip setq
+	(forward-sexp)
+	;; skip ellama--current-session
+	(forward-sexp)
+	;; skip space
+	(forward-char)
+	;; remove all above
+	(kill-region (point-min) (point))
+	(goto-char (point-max))
+	;; remove ")"
+	(delete-char -1)
+	;; save session in new format
+	(save-buffer)
+	(goto-char (point-min))))
     (with-current-buffer buffer
-      (eval (read session-buffer))
+      (setq ellama--current-session (read session-buffer))
       (setq ellama--current-session-id (ellama-session-id ellama--current-session))
       (puthash (ellama-session-id ellama--current-session)
-	       buffer ellama--active-sessions))
+	       buffer ellama--active-sessions)
+      (ellama-session-mode +1))
     (kill-buffer session-buffer)
     (display-buffer buffer)))
 
@@ -433,9 +559,7 @@ PROMPT is a variable contains last prompt in this session."
 	 (buffer (ellama-get-session-buffer id))
 	 (file (buffer-file-name buffer))
 	 (session-file (ellama--get-session-file-name file)))
-    (with-current-buffer buffer
-      (ellama--session-deactivate)
-      (kill-buffer))
+    (kill-buffer buffer)
     (delete-file file t)
     (delete-file session-file t)))
 
@@ -500,6 +624,9 @@ strings before they're inserted into the BUFFER.
 
 :session SESSION -- SESSION is a ellama conversation session.
 
+:ephemeral-session BOOL -- if BOOL is set session will not be saved to named
+file by default.
+
 :on-error ON-ERROR -- ON-ERROR a function that's called with an error message on
 failure (with BUFFER current).
 
@@ -529,6 +656,7 @@ when the request completes (with BUFFER current)."
 				 (llm-make-simple-chat-prompt prompt)))
 		       (llm-make-simple-chat-prompt prompt))))
     (with-current-buffer buffer
+      (ellama-request-mode +1)
       (let* ((start (make-marker))
 	     (end (make-marker))
 	     (insert-text
@@ -544,15 +672,17 @@ when the request completes (with BUFFER current)."
                     (when (pcase ellama-fill-paragraphs
                             ((cl-type function) (funcall ellama-fill-paragraphs))
                             ((cl-type boolean) ellama-fill-paragraphs)
-                            ((cl-type list) (apply #'derived-mode-p
-						   ellama-fill-paragraphs)))
+                            ((cl-type list) (and (apply #'derived-mode-p
+							ellama-fill-paragraphs)
+						 (not (equal major-mode 'org-mode)))))
                       (fill-region start (point)))
 		    (goto-char pt))
 		  (when-let ((ellama-auto-scroll)
 			     (window (get-buffer-window buffer)))
 		    (with-selected-window window
 		      (goto-char (point-max))
-		      (recenter -1)))))))
+		      (recenter -1)))
+		  (undo-amalgamate-change-group ellama--change-group)))))
 	(setq ellama--change-group (prepare-change-group))
 	(activate-change-group ellama--change-group)
 	(set-marker start point)
@@ -567,24 +697,27 @@ when the request completes (with BUFFER current)."
 				  (lambda (text)
 				    (funcall insert-text text)
 				    (with-current-buffer buffer
-				      (undo-amalgamate-change-group ellama--change-group)
 				      (accept-change-group ellama--change-group)
 				      (spinner-stop)
 				      (funcall donecb text)
-				      (setq ellama--current-request nil)))
+				      (setq ellama--current-request nil)
+				      (ellama-request-mode -1)))
 				  (lambda (_ msg)
 				    (with-current-buffer buffer
 				      (cancel-change-group ellama--change-group)
 				      (spinner-stop)
 				      (funcall errcb msg)
-				      (setq ellama--current-request nil)))))))))
+				      (setq ellama--current-request nil)
+				      (ellama-request-mode -1)))))))))
 
 (defun ellama-chat-done (text)
   "Chat done.
 Will call `ellama-chat-done-callback' on TEXT."
   (save-excursion
     (goto-char (point-max))
-    (insert "\n\n"))
+    (insert "\n\n")
+    (when ellama-session-auto-save
+      (save-buffer)))
   (when ellama-chat-done-callback
     (funcall ellama-chat-done-callback text)))
 
@@ -627,7 +760,9 @@ If CREATE-SESSION set, creates new session even if there is an active session."
 		ellama-nick-prefix " " ellama-assistant-nick ":\n")
 	(ellama-stream prompt
 		       :session session
-		       :on-done #'ellama-chat-done)))))
+		       :on-done #'ellama-chat-done
+		       :filter (when (derived-mode-p 'org-mode)
+				 #'ellama--translate-markdown-to-org-filter))))))
 
 ;;;###autoload
 (defun ellama-ask-about ()
@@ -673,9 +808,16 @@ If CREATE-SESSION set, creates new session even if there is an active session."
   (let* ((buffer-name (ellama-generate-name ellama-provider real-this-command prompt))
 	 (buffer (get-buffer-create (if (get-buffer buffer-name)
 					(make-temp-name (concat buffer-name " "))
-				      buffer-name))))
+				      buffer-name)))
+	 filter)
+    (with-current-buffer buffer
+      (funcall ellama-major-mode)
+      (when (derived-mode-p 'org-mode)
+	(setq filter 'ellama--translate-markdown-to-org-filter)))
     (display-buffer buffer)
-    (ellama-stream prompt :buffer buffer (point-min))))
+    (ellama-stream prompt
+		   :buffer buffer
+		   :filter filter)))
 
 ;;;###autoload
 (defun ellama-translate ()
